# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Top-level package for the Q-CTRL Visualizer.

The public API of this package consists only of the objects exposed through this top-level package.
Direct access to sub-modules is not officially supported, so may be affected by
backwards-incompatible changes without notice.
"""
from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

from .bloch import (
    display_bloch_sphere,
    display_bloch_sphere_from_bloch_vectors,
    display_bloch_sphere_from_density_matrices,
)
from .confidence_ellipses import plot_confidence_ellipses
from .controls import (
    plot_controls,
    plot_sequences,
    plot_smooth_controls,
)
from .cost_history import plot_cost_history
from .filter_functions import plot_filter_functions
from .histogram import plot_bitstring_probabilities_histogram
from .populations import plot_populations
from .style import (
    QCTRL_STYLE_COLORS,
    get_qctrl_style,
)
from .wigner_function import plot_wigner_function

__all__ = [
    "QCTRL_STYLE_COLORS",
    "display_bloch_sphere",
    "display_bloch_sphere_from_bloch_vectors",
    "display_bloch_sphere_from_density_matrices",
    "get_qctrl_style",
    "plot_confidence_ellipses",
    "plot_controls",
    "plot_cost_history",
    "plot_filter_functions",
    "plot_populations",
    "plot_bitstring_probabilities_histogram",
    "plot_sequences",
    "plot_smooth_controls",
    "plot_wigner_function",
]

__version__ = "4.2.0"



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Functions for plotting confidence ellipses.
"""
from __future__ import annotations

from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from qctrlcommons.preconditions import check_argument

from .style import (
    FIG_HEIGHT,
    QCTRL_STYLE_COLORS,
    qctrl_style,
)
from .utils import (
    create_axes,
    figure_as_kwarg_only,
    get_units,
)


@qctrl_style()
@figure_as_kwarg_only
def plot_confidence_ellipses(
    ellipse_matrix: np.ndarray,
    estimated_parameters: np.ndarray,
    actual_parameters: Optional[np.ndarray] = None,
    parameter_names: Optional[list[str]] = None,
    parameter_units: str | list[str] = "Hz",
    *,
    figure: plt.Figure,
):
    """
    Create an array of confidence ellipse plots.

    From an (N,N) matrix transformation and N estimated parameters,
    plots the confidence ellipse for each pair of parameters.

    Parameters
    ----------
    ellipse_matrix : np.ndarray
        The square matrix which transforms a unit hypersphere in an
        N-dimensional space into a hyperellipse representing the confidence
        region. Must be of shape (N, N), with N > 1.
    estimated_parameters : np.ndarray
        The values of the estimated parameters. Must be of shape (N,).
    actual_parameters : np.ndarray, optional
        The actual values of the estimated parameters.
        If you provide these, they're plotted alongside the ellipses and
        estimated parameters. Must be of shape (N,).
    parameter_names : list[str], optional
        The name of each parameter, to be used as axes labels.
        If provided, it must be of length N. If not provided,
        the axes are labelled "Parameter 0", "Parameter 1", ...
    parameter_units : str or list[str], optional
        The units of each parameter. You can provide a list of strings with
        the units of each parameter, or a single string if all parameters have
        the same units. Defaults to "Hz".
    figure : matplotlib.figure.Figure, optional
        A matplotlib Figure in which to place the plots.
        If passed, its dimensions and axes will be overridden.
    """

    ellipse_matrix = np.asarray(ellipse_matrix)
    estimated_parameters = np.asarray(estimated_parameters)
    if actual_parameters is not None:
        actual_parameters = np.asarray(actual_parameters)

    check_argument(
        len(estimated_parameters.shape) == 1 and len(estimated_parameters) > 1,
        "The estimated parameters must be a 1D array containing at least two parameters.",
        {"estimated_parameters": estimated_parameters},
        extras={"estimated_parameters.shape": estimated_parameters.shape},
    )

    parameter_count = len(estimated_parameters)

    check_argument(
        ellipse_matrix.shape == (parameter_count, parameter_count),
        "The ellipse_matrix must be a square 2D array, "
        "with the same length as the estimated parameters.",
        {"ellipse_matrix": ellipse_matrix},
        extras={
            "ellipse_matrix.shape": ellipse_matrix.shape,
            "len(estimated_parameters)": parameter_count,
        },
    )

    if actual_parameters is not None:
        check_argument(
            actual_parameters.shape == (parameter_count,),
            "If passed, the actual parameters must be a 1D array "
            "with the same shape as estimated parameters.",
            {"actual_parameters": actual_parameters},
            extras={
                "actual_parameters.shape": actual_parameters.shape,
                "estimated_parameters.shape": estimated_parameters.shape,
            },
        )

    if parameter_names is not None:
        check_argument(
            len(parameter_names) == parameter_count,
            "If passed, the parameter names must be a list "
            "with the same length as the estimated parameters.",
            {"parameter_names": parameter_names},
            extras={
                "len(parameter_names)": len(parameter_names),
                "len(estimated_parameters)": parameter_count,
            },
        )

    else:
        parameter_names = [f"Parameter {k}" for k in range(parameter_count)]

    if isinstance(parameter_units, str):
        parameter_units = [parameter_units] * parameter_count
    else:
        check_argument(
            len(parameter_units) == parameter_count,
            "The parameter units must be either a string or a list "
            "with the same length as the estimated parameters.",
            {"parameter_units": parameter_units},
            extras={
                "len(parameter_units)": len(parameter_units),
                "len(estimated_parameters)": parameter_count,
            },
        )

    # Set the N (N-1) / 2 plots in a 2D grid of axes.
    if parameter_count % 2 == 0:
        plot_count_x = parameter_count - 1
        plot_count_y = parameter_count // 2
    else:
        plot_count_x = parameter_count
        plot_count_y = (parameter_count - 1) // 2

    axes_array = create_axes(
        figure, plot_count_x, plot_count_y, FIG_HEIGHT, FIG_HEIGHT
    ).flatten()

    # Create pairs of indices with all possible parameter pairings.
    index_1_list, index_2_list = np.triu_indices(parameter_count, k=1)

    for axes, index_1, index_2 in zip(axes_array, index_1_list, index_2_list):

        # Obtain points representing the correct parameters and their estimates.
        estimated_dot = estimated_parameters[[index_1, index_2]]

        # Obtain coordinates for a circle.
        theta = np.linspace(0, 2 * np.pi, 101)
        circle_coordinates = np.array([np.cos(theta), np.sin(theta)])

        # Define matrix that transforms circle coordinates into ellipse coordinates.
        coordinate_change = ellipse_matrix[
            np.ix_([index_1, index_2], [index_1, index_2])
        ]
        ellipse = coordinate_change @ circle_coordinates + estimated_dot[:, None]
        scale_x, units_x = get_units(ellipse[0])
        scale_y, units_y = get_units(ellipse[1])
        scale = np.array([scale_x, scale_y])

        # Define labels of the axes.
        axes.set_xlabel(
            f"{parameter_names[index_1]} ({units_x}{parameter_units[index_1]})",
            labelpad=0,
        )
        axes.set_ylabel(
            f"{parameter_names[index_2]} ({units_y}{parameter_units[index_2]})",
            labelpad=0,
        )

        # Plot estimated parameters.
        estimated_dot = estimated_dot / scale
        axes.plot(
            *estimated_dot, "o", label="Estimated parameters", c=QCTRL_STYLE_COLORS[0]
        )

        # Plot confidence ellipse.
        ellipse = ellipse / scale[:, None]
        axes.plot(*ellipse, "--", label="Confidence region", c=QCTRL_STYLE_COLORS[0])

        # Plot actual parameters (if available).
        if actual_parameters is not None:
            actual_dot = actual_parameters[[index_1, index_2]] / scale
            axes.plot(
                *actual_dot, "o", label="Actual parameters", c=QCTRL_STYLE_COLORS[1]
            )

    # Create legends.
    handles, labels = axes_array[0].get_legend_handles_labels()
    figure.legend(
        handles=handles, labels=labels, loc="center", bbox_to_anchor=(0.5, 0.95), ncol=3
    )



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Functions for handling Q-CTRL styling.
"""

from __future__ import annotations

from contextlib import contextmanager
from typing import Any

from cycler import cycler
from matplotlib import rcParams
from matplotlib.style import context

BORDER_COLOR = "#D8E0E9"
TEXT_COLOR = "#6C5C71"

FIG_WIDTH = 10.0
FIG_HEIGHT = 5.0

QCTRL_STYLE_COLORS = [
    "#680CE9",
    "#D84144",
    "#32A4A8",
    "#A2A933",
    "#D84190",
    "#4177D8",
    "#D6722F",
    "#D543ED",
    "#32A857",
]


def get_qctrl_style() -> dict[str, Any]:
    """
    Return a dictionary representing the Q-CTRL styling in Matplotlib format.

    The returned dictionary is suitable for passing to Matplotlib functions such as
    ``matplotlib.style.use`` or ``matplotlib.style.context``.

    Returns
    -------
    dict
        The dictionary representing the Q-CTRL style.
    """
    style: dict[str, Any] = {}

    # Set text color.
    style["text.color"] = TEXT_COLOR
    style["xtick.color"] = TEXT_COLOR
    style["ytick.color"] = TEXT_COLOR
    style["axes.labelcolor"] = TEXT_COLOR

    # Set font to Roboto.
    style["font.sans-serif"] = ["Roboto"] + rcParams["font.sans-serif"]

    # Set figure size.
    style["figure.figsize"] = [FIG_WIDTH, FIG_HEIGHT]

    # Set borders and their color.
    style["axes.spines.left"] = True
    style["axes.spines.right"] = True
    style["axes.spines.top"] = True
    style["axes.spines.bottom"] = True
    style["axes.edgecolor"] = BORDER_COLOR
    style["legend.edgecolor"] = BORDER_COLOR
    style["lines.color"] = BORDER_COLOR

    # Set font sizes.
    style["font.size"] = 14
    style["legend.title_fontsize"] = 14
    style["legend.fontsize"] = 14
    style["figure.titlesize"] = 16
    style["axes.titlesize"] = 14
    style["axes.labelsize"] = 14
    style["xtick.labelsize"] = 12
    style["ytick.labelsize"] = 12

    # Set padding around labels.
    style["axes.labelpad"] = 15.0

    # Set background color white.
    style["figure.facecolor"] = "white"

    # Set cycle of colors for lines.
    style["axes.prop_cycle"] = cycler(color=QCTRL_STYLE_COLORS)

    # Set line width.
    style["lines.linewidth"] = 2

    return style


@contextmanager
def qctrl_style():
    """
    A ``ContextDecorator`` that enables the Q-CTRL Matplotlib styling.

    The returned object can act as a decorator::

        @qctrl_style()
        def plot(*args):
            # Matplotlib calls made in this function will use Q-CTRL styling.
            pass

    The returned object can also act as a context manager::

        with qctrl_style():
            # Matplotlib calls made in this context will use Q-CTRL styling.
            pass
    """
    with context(get_qctrl_style()) as ctx:
        yield ctx



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.


"""
Function for plotting the populations.
"""
from __future__ import annotations

from collections import namedtuple

import matplotlib.pyplot as plt
import numpy as np
from qctrlcommons.preconditions import check_argument

from .style import (
    FIG_HEIGHT,
    FIG_WIDTH,
    qctrl_style,
)
from .utils import (
    figure_as_kwarg_only,
    get_units,
)


@qctrl_style()
@figure_as_kwarg_only
def plot_populations(
    sample_times: np.ndarray, populations: dict, *, figure: plt.Figure
):
    """
    Create a plot of the specified populations.

    Parameters
    ----------
    sample_times : np.ndarray
        The 1D array of times in seconds at which the populations have been sampled.
    populations : dict
        The dictionary of populations to plot, of the form
        ``{"label_1": population_values_1, "label_2": population_values_2, ...}``.
        Each `population_values_n` is a 1D array of population values with the same
        length as `sample_times` and `label_n` is its label.
        Population values must lie between 0 and 1.
    figure : matplotlib.figure.Figure, optional
        A matplotlib Figure in which to place the plots.
        If passed, its dimensions and axes will be overridden.
    """

    population_data = _create_population_data(sample_times, populations)

    figure.set_figwidth(FIG_WIDTH)
    figure.set_figheight(FIG_HEIGHT)

    axes = figure.subplots(nrows=1, ncols=1)

    scale, prefix = get_units(sample_times)
    for data in population_data:
        axes.plot(sample_times / scale, data.values, label=data.label)

    axes.set_xlabel(f"Time ({prefix}s)")
    axes.set_ylabel("Probability")

    axes.legend()


_PopulationData = namedtuple("_PopulationData", ["values", "label"])


def _create_population_data(
    sample_times: np.ndarray, populations: dict
) -> list[_PopulationData]:
    """
    Validate inputs and create a list of _PopulationData objects.

    Parameters
    ----------
    sample_count : np.ndarray
        The times at which the populations have been sampled.
    populations : dict
        The populations to plot.

    Returns
    -------
    list[_PopulationData]
        A list of _PopulationData.
    """

    def safe_less_than(arr, bound):
        """
        Returns False if any of the elements of arr are less than bound (within rounding error).
        Returns True otherwise.
        """
        return np.any((1 - np.isclose(arr, bound)) * (arr < bound))

    def safe_greater_than(arr, bound):
        """
        Returns False if any of the elements of arr are greater than bound (within rounding error).
        Returns True otherwise.
        """
        return np.any((1 - np.isclose(arr, bound)) * (arr > bound))

    check_argument(
        isinstance(sample_times, np.ndarray) and len(sample_times.shape) == 1,
        "The sample times must be a 1D array.",
        {"sample_times": sample_times},
    )
    check_argument(
        isinstance(populations, dict),
        "The populations must be a dictionary.",
        {"populations": populations},
    )

    sample_count = len(sample_times)

    plot_data = []
    for label, pop in populations.items():
        check_argument(
            isinstance(pop, (list, np.ndarray)),
            "Each element in the dictionary of populations must be an array or a list.",
            {"populations": populations},
            extras={f"populations[{label}]": pop},
        )
        check_argument(
            not safe_less_than(np.asarray(pop), 0)
            and (not safe_greater_than(np.asarray(pop), 1)),
            "Population values must lie between 0 and 1.",
            {"populations": populations},
            extras={f"populations[{label}]": pop},
        )
        check_argument(
            len(pop) == sample_count,
            "The number of population values must match the number of sample times.",
            {"sample_times": sample_times, "populations": populations},
            extras={
                f"len(populations[{label}])": len(pop),
                "len(sample_times)": sample_count,
            },
        )
        plot_data.append(_PopulationData(np.asarray(pop), label))

    return plot_data



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Functions for plotting cost vs iterations.
"""

from __future__ import annotations

from typing import Optional

import matplotlib.pyplot as plt
from qctrlcommons.preconditions import check_argument

from .style import (
    FIG_HEIGHT,
    FIG_WIDTH,
    qctrl_style,
)
from .utils import figure_as_kwarg_only


@qctrl_style()
@figure_as_kwarg_only
def plot_cost_history(
    cost_histories: list,
    labels: Optional[list[str]] = None,
    y_axis_log: bool = False,
    initial_iteration: int = 1,
    *,
    figure: plt.Figure,
):
    """
    Create a plot of the cost against iterations for either a single cost history or
    a set of cost histories.

    Parameters
    ----------
    cost_histories : list
        The values of the cost histories.
        Must be either a list of a single cost history or a list of several cost histories,
        where each individual cost history is a list.
        For example, a single cost history can be passed
        (``cost_histories = [0.1, 0.05, 0.02]``) or multiple cost histories
        (``cost_histories = [[0.1, 0.05, 0.02], [0.23, 0.2, 0.14, 0.1, 0.04, 0.015]]``).
    labels : list[str], optional
        The labels corresponding to the individual cost histories in `cost_histories`.
        If you provide this, it must be the same length as `cost_histories`.
    y_axis_log : bool, optional
        Whether the y-axis is log-scale.
        Defaults to False.
    initial_iteration : int, optional
        Where the iteration count on the x-axis starts from.
        This is useful if you want to include the initial cost—before optimization—at
        iteration 0 or if you pass cost histories that start at a later iteration.
        Defaults to 1.
    figure : matplotlib.figure.Figure, optional
        A matplotlib Figure in which to place the plots.
        If passed, its dimensions and axes will be overridden.
    """

    figure.set_figwidth(FIG_WIDTH)
    figure.set_figheight(FIG_HEIGHT)

    axs = figure.subplots(nrows=1, ncols=1)

    check_argument(
        isinstance(cost_histories, list),
        "The cost histories must be in a list.",
        {"cost_histories": cost_histories},
    )

    if not all(isinstance(history, list) for history in cost_histories):
        cost_histories = [cost_histories]

    if labels is not None:
        check_argument(
            len(cost_histories) == len(labels),
            "If passed, the labels must have the same length as the list of cost histories.",
            {"labels": labels, "cost_histories": cost_histories},
        )

        for cost_history, label in zip(cost_histories, labels):
            axs.plot(
                range(initial_iteration, initial_iteration + len(cost_history)),
                cost_history,
                label=label,
            )
        axs.legend(loc="upper left", bbox_to_anchor=(1, 1))
    else:
        for cost_history in cost_histories:
            axs.plot(
                range(initial_iteration, initial_iteration + len(cost_history)),
                cost_history,
            )

    axs.set_xlabel("Iteration")
    axs.set_ylabel("Cost")

    if y_axis_log:
        axs.set_yscale("log")



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.


"""
Function for plotting histograms.
"""
from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_non_negative_scalar,
)

from .style import (
    FIG_HEIGHT,
    FIG_WIDTH,
    qctrl_style,
)
from .utils import figure_as_kwarg_only


def _validate_bitstring_counts(bitstring_counts: dict[str, dict[str, int]]) -> int:
    """
    Validate the input arguments of `plot_bitstring_probabilities_histogram`.

    Parameters
    ----------
    bitstring_counts : dict[str, dict[str, int]]
        Nested dictionaries of experiments with the resulting bitstring counts to plot.

    Returns
    -------
    int
        The length of the bitstrings.
    """

    check_argument(
        isinstance(bitstring_counts, dict),
        "The input bitstring counts must be in a dictionary.",
        {"bitstring_counts": bitstring_counts},
    )

    for experiment, counts in bitstring_counts.items():
        check_argument(
            isinstance(bitstring_counts[experiment], dict),
            "The counts must be dictionaries.",
            {"bitstring_counts": bitstring_counts},
            extras={f"bitstring_counts[{experiment}]": counts},
        )
        check_argument(
            len(bitstring_counts[experiment]) > 0,
            "The inner count dictionaries must not be empty.",
            {"bitstring_counts": bitstring_counts},
            extras={f"bitstring_counts[{experiment}]": bitstring_counts[experiment]},
        )

    first_key = next(iter(bitstring_counts))
    bitstring_length = len(next(iter(bitstring_counts[first_key])))

    for experiment, counts in bitstring_counts.items():
        for bitstring, count in counts.items():
            check_argument_non_negative_scalar(
                count, f'bitstring_counts["{experiment}"]["{bitstring}"]'
            )
            check_argument(
                len(bitstring) == bitstring_length,
                "The length of all bitstrings must be the same.",
                {"bitstring_counts": bitstring_counts},
                extras={"first bitstring length": bitstring_length},
            )

    return bitstring_length


@qctrl_style()
@figure_as_kwarg_only
def plot_bitstring_probabilities_histogram(
    bitstring_counts: dict[str, dict[str, int]],
    show_legend: bool = True,
    display_all_bitstrings: bool = False,
    *,
    figure: plt.Figure,
) -> None:
    """
    Create a histogram of the specified counts.

    Parameters
    ----------
    bitstring_counts : dict[str, dict[str,int]]
        Nested dictionaries of experiments with the resulting bitstring counts to plot, of the form
        ``{"Fire Opal": {"00...0": count_0_1 , ..., "11...1": count_0_2n}, ...,
        "IBM": {"00...0": count_1_1, ..., "11...1": count_1_2n}}``.
        The keys of the outer dictionary will be used for the plot legend,
        and the inner dictionaries keys represent the bitstrings whose counts will be plotted.
        The counts must be nonnegative and the length of all bitstring across all
        experiments must be the same.
    show_legend : bool, optional
        Whether to add a legend to the plot. Defaults to True.
    display_all_bitstrings : bool, optional
        Whether to display all possible bitstrings, including those not
        present in `bitstring_counts`. Defaults to False.
    figure : matplotlib.figure.Figure, optional
        A matplotlib Figure in which to place the plots.
        If passed, its dimensions and axes will be overridden.
    """

    bitstring_length = _validate_bitstring_counts(bitstring_counts)

    if display_all_bitstrings:
        sorted_bitstrings = [
            np.binary_repr(i, bitstring_length) for i in range(2**bitstring_length)
        ]
    else:
        # Sort the bitstring labels.
        sorted_bitstrings = sorted(set().union(*list(bitstring_counts.values())))

    figure.set_figwidth(FIG_WIDTH)
    figure.set_figheight(FIG_HEIGHT)

    axes = figure.subplots(nrows=1, ncols=1)

    default_bar_width = 0.8
    bar_width = default_bar_width / len(bitstring_counts)

    x_axis = np.arange(len(sorted_bitstrings))

    for bar_shift_index, (experiment, counts) in enumerate(bitstring_counts.items()):

        counts_array = np.zeros(len(sorted_bitstrings))
        for index, bitstring in enumerate(sorted_bitstrings):
            counts_array[index] = counts.get(bitstring, 0)

        total_counts = np.sum(counts_array)

        axes.bar(
            x_axis + bar_shift_index * bar_width,
            counts_array / total_counts,
            width=bar_width,
            label=experiment,
        )

    axes.set_xticks(
        x_axis + 0.5 * (default_bar_width - bar_width), sorted_bitstrings, rotation=90
    )
    axes.set_xlabel("Bitstrings")
    axes.set_ylabel("Probability")
    if show_legend:
        axes.legend(loc="upper left", bbox_to_anchor=(1, 1))



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Utility functions for plotting.
"""

from __future__ import annotations

from functools import wraps
from inspect import (
    Parameter,
    Signature,
    signature,
)
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import ScalarFormatter
from qctrlcommons.preconditions import check_argument


def get_units(values) -> tuple[float, str]:
    """
    Calculate the units to be used for plotting the given range of values.

    Specifically, returns a tuple (scaling factor, prefix), for example (1000, 'k') or (0.001, 'm').
    The values should be divided by the first element, and the unit label prepended with the second
    element. If the values are zero the scaling factor is 1.
    """
    prefixes = {
        -24: "y",
        -21: "z",
        -18: "a",
        -15: "f",
        -12: "p",
        -9: "n",
        -6: "\N{MICRO SIGN}",
        -3: "m",
        0: "",
        3: "k",
        6: "M",
        9: "G",
        12: "T",
        15: "P",
        18: "E",
        21: "Z",
        24: "Y",
    }
    # We apply a simple algorithm: get the element with largest magnitude, then map according to
    # [0.001, 1) -> 0.001x/milli,
    # [1, 1000) -> no scaling,
    # [1000, 1e6) -> 1000x/kilo,
    # and so on.
    max_value = max(np.abs(values))
    exponent = float(
        3
        * np.floor_divide(np.log10(max_value, out=np.zeros(1), where=max_value > 0), 3)
    )
    # Clip the scaling to the allowable range.
    exponent_clipped = np.clip(exponent, -24, 24)
    return 10**exponent_clipped, prefixes[exponent_clipped]


def create_axes(
    figure: plt.Figure,
    count_x: int,
    count_y: int,
    width: float,
    height: float,
    share_x: bool = False,
    share_y: bool = False,
) -> np.ndarray:
    """
    Create a set of axes with default axis labels and axis formatting.

    Parameters
    ----------
    figure : matplotlib.figure.Figure
        The matplotlib Figure in which the axes should be placed.
        Its dimensions will be overridden by this method.
    count_x : int
        The number of axes to create horizontally.
    count_y : int
        The number of axes to create vertically.
    width : float
        The width (in inches) for each axes.
    height : float
        The height (in inches) for each axes.
    share_x : bool, optional
        Whether the axes share the x-axis.
        Defaults to False.
    share_y : bool, optional
        Whether the axes share the y-axis.
        Defaults to False.

    Returns
    -------
    np.ndarray
        A 2D array of Axes objects.
    """
    figure.set_figheight(height * count_y)
    figure.set_figwidth(width * count_x)
    figure.subplots_adjust(hspace=0.2, wspace=0.25)

    axes_array = figure.subplots(
        nrows=count_y, ncols=count_x, sharex=share_x, sharey=share_y, squeeze=False
    )

    # Set axis formatting.
    for axes in axes_array.flat:
        axes.yaxis.set_major_formatter(ScalarFormatter())
        axes.xaxis.set_major_formatter(ScalarFormatter())

    return axes_array


def figure_as_kwarg_only(func):
    """
    Decorator to mark functions which take a figure as kwarg-only parameter.
    The decorator:
        - checks that the first positional argument passed
            to the decorated function is not a Figure.
        - if no kwarg figure is passed, create one
            and pass it to the decorated function.
        - if a kwarg figure is passed, remove all of its axes.
    """

    @wraps(func)
    def new_func(*args, **kwargs):
        if len(args) > 0:
            check_argument(  # Remove 2023/01/01
                not isinstance(args[0], plt.Figure),
                "The figure parameter is now optional. "
                "You can omit it or pass it as a keyword parameter.",
                {func.__code__.co_varnames[0]: args[0]},
            )

        figure = kwargs.get("figure")
        if figure is None:
            # If a figure is not passed, create one.
            kwargs["figure"] = plt.figure()
        else:
            # If a figure is passed, remove all of its axes.
            for axes in figure.get_axes():
                axes.remove()

        return func(*args, **kwargs)

    # Update decorated function signature and annotations making the figure parameter optional.
    sig_params = dict(signature(new_func).parameters)
    sig_params["figure"] = Parameter(
        "figure", Parameter.KEYWORD_ONLY, default=None, annotation=Optional[plt.Figure]
    )
    # Ignoring type due to https://github.com/python/mypy/issues/12472
    new_func.__signature__ = Signature(parameters=list(sig_params.values()))  # type: ignore
    new_func.__annotations__["figure"] = Optional[plt.Figure]

    return new_func



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Functions for plotting filter functions.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from qctrlcommons.preconditions import check_argument

from .style import (
    FIG_HEIGHT,
    FIG_WIDTH,
    qctrl_style,
)
from .utils import figure_as_kwarg_only


@qctrl_style()
@figure_as_kwarg_only
def plot_filter_functions(
    filter_functions: dict,
    *,
    figure: plt.Figure,
):
    """
    Create a plot of the specified filter functions.

    Parameters
    ----------
    filter_functions : dict
        The dictionary of filter functions to plot. The keys should be the names of the filter
        functions, and the values represent the filter filter by either a dictionary with the
        keys 'frequencies', 'inverse_powers', and optional 'inverse_power_uncertainties' or
        a list of samples that each contain a dictionary with keys 'frequency', 'inverse_power',
        and optional 'inverse_power_uncertainty'. The frequency must be in Hertz an the inverse
        power and its uncertainty in seconds.

        For the list of samples the key 'inverse_power_precision' can be used instead of
        'inverse_power_uncertainty'. If both are provided then the value corresponding to
        'inverse_power_uncertainty' is used.

        If the uncertainty of an inverse power is provided, it must be non-negative.

        For example, the following is a valid ``filter_functions`` input::

            {
             "Primitive": {
                "frequencies": [0.0, 1.0, 2.0],
                "inverse_powers": [15., 12., 3.],
                "inverse_power_uncertainties": [0., 0., 0.2],
             },
             "CORPSE": [
                {"frequency": 0.0, "inverse_power": 10.},
                {"frequency": 0.5, "inverse_power": 8.5},
                {"frequency": 1.0, "inverse_power": 5., "inverse_power_uncertainty": 0.1},
                {"frequency": 1.5, "inverse_power": 2.5},
             ],
            }
    figure : matplotlib.figure.Figure, optional
        A matplotlib Figure in which to place the plot.
        If passed, its dimensions and axes will be overridden.
    """

    check_argument(
        filter_functions,
        "At least one filter function must be provided.",
        {"filter_functions": filter_functions},
    )

    figure.set_figwidth(FIG_WIDTH)
    figure.set_figheight(FIG_HEIGHT)

    axes = figure.subplots(nrows=1, ncols=1)

    for name, filter_function in filter_functions.items():
        if isinstance(filter_function, list):
            frequencies, inverse_powers, inverse_power_uncertainties = np.array(
                list(
                    zip(
                        *[
                            (
                                sample["frequency"],
                                sample["inverse_power"],
                                sample["inverse_power_uncertainty"]
                                if "inverse_power_uncertainty" in sample
                                else sample.get("inverse_power_precision", 0.0),
                            )
                            for sample in filter_function
                        ]
                    )
                )
            )
        else:
            check_argument(
                isinstance(filter_function, dict),
                "Each filter function must either be a list or a dictionary.",
                {"filter_functions": filter_functions},
                extras={"filter_function": filter_function},
            )
            check_argument(
                ("frequencies" in filter_function)
                and ("inverse_powers" in filter_function),
                "Each filter function dictionary must contain `frequencies` and"
                " `inverse_powers` keys.",
                {"filter_functions": filter_functions},
                extras={"filter_function": filter_function},
            )
            frequencies = np.asarray(filter_function["frequencies"])
            inverse_powers = np.asarray(filter_function["inverse_powers"])
            inverse_power_uncertainties = (
                np.asarray(filter_function["inverse_power_uncertainties"])
                if "inverse_power_uncertainties" in filter_function
                else np.zeros(len(frequencies))
            )

        check_argument(
            np.all(inverse_power_uncertainties >= 0.0),
            "Uncertainties must all be non-negative in filter functions.",
            {"filter_functions": filter_functions},
            extras={"filter_function": filter_function},
        )

        inverse_powers_upper = inverse_powers + inverse_power_uncertainties
        inverse_powers_lower = inverse_powers - inverse_power_uncertainties

        lines = axes.plot(frequencies, inverse_powers, label=name)
        axes.fill_between(
            frequencies,
            inverse_powers_lower,
            inverse_powers_upper,
            alpha=0.35,
            hatch="||",
            facecolor="none",
            edgecolor=lines[0].get_color(),
            linewidth=0,
        )

    axes.legend()

    axes.set_xscale("log")
    axes.set_yscale("log")

    axes.autoscale(axis="x", tight=True)

    axes.set_xlabel("Frequency (Hz)")
    axes.set_ylabel("Inverse power (s)")



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Functions for plotting control pulses.
"""

from __future__ import annotations

from collections import namedtuple
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from qctrlcommons.preconditions import check_argument

from .style import (
    FIG_WIDTH,
    qctrl_style,
)
from .utils import (
    create_axes,
    figure_as_kwarg_only,
    get_units,
)


@qctrl_style()
@figure_as_kwarg_only
def plot_controls(
    controls: dict,
    polar: bool = True,
    smooth: bool = False,
    unit_symbol: str = "Hz",
    two_pi_factor: bool = True,
    *,
    figure: plt.Figure,
):
    """
    Create a plot of the specified controls.

    Parameters
    ----------
    controls : dict
        The dictionary of controls to plot. The keys should be the names of the controls, and the
        values represent the pulse by either (1) a dictionary with the 'durations' and 'values' for
        that control, or (2) a list of segments, each a dictionary with 'duration' and 'value' keys.
        The durations must be in seconds and the values (possibly complex) in the units specified by
        `unit_symbol`.
        For example, the following two ``controls`` inputs would be valid (and equivalent)::

            controls={
              "Clock": {"durations": [1.0, 1.0, 2.0], "values": [-0.5, 0.5, -1.5]},
              "Microwave": {"durations": [0.5, 1.0], "values": [0.5 + 1.5j, 0.2 - 0.3j]},
            }

            controls={
              "Clock": [
                {"duration": 1.0, "value": -0.5},
                {"duration": 1.0, "value": 0.5},
                {"duration": 2.0, "value": -1.5},
              ],
              "Microwave": [
                {"duration": 0.5, "value": 0.5 + 1.5j},
                {"duration": 1.0, "value": 0.2 - 0.3j},
              ],
            }

    polar : bool, optional
        The mode of the plot when the values appear to be complex numbers.
        Plot magnitude and angle in two figures if set to True, otherwise plot I and Q
        in two figures. Defaults to True.
    smooth : bool, optional
        Whether to plot the controls as samples joined by straight lines, rather than as
        piecewise-constant segments. Defaults to False.
    unit_symbol : str, optional
        The symbol of the unit to which the controls values correspond. Defaults to "Hz".
    two_pi_factor : bool, optional
        Whether the values of the controls should be divided by 2π in the plots.
        Defaults to True.
    figure : matplotlib.figure.Figure, optional
        A matplotlib Figure in which to place the plots.
        If passed, its dimensions and axes will be overridden.
    """

    boulder_opal_advice = (
        "If you are attempting to plot controls from Boulder Opal, you can obtain"
        " controls in the correct format by using Pwcs as the output nodes."
    )

    plots_data: list[Any] = []

    check_argument(
        isinstance(controls, dict),
        f"The controls must be in a dictionary. {boulder_opal_advice}",
        {"controls": controls},
    )

    for name, control in controls.items():
        # Process each control depending whether it's passed as a list of segments...
        if isinstance(control, list):
            durations = []
            values = []
            for segment in control:
                check_argument(
                    isinstance(segment, dict),
                    f"Each control segment must be a dictionary. {boulder_opal_advice}",
                    {"controls": controls},
                    extras={"segment": segment},
                )

                check_argument(
                    ("duration" in segment) and ("value" in segment),
                    "Each control segment dictionary must contain `duration` and"
                    f" `value` keys. {boulder_opal_advice}",
                    {"controls": controls},
                    extras={"segment": segment},
                )

                durations.append(segment["duration"])
                values.append(segment["value"])
        # ...or a durations/values dictionary.
        else:
            check_argument(
                isinstance(control, dict),
                f"Each control must either be a list or a dictionary. {boulder_opal_advice}",
                {"controls": controls},
                extras={"control": control},
            )

            check_argument(
                ("durations" in control) and ("values" in control),
                "Each control dictionary must contain `durations` and"
                f" `values` keys. {boulder_opal_advice}",
                {"controls": controls},
                extras={"control": control},
            )

            durations = control["durations"]
            values = control["values"]
        # Create plot data for the control.
        plots_data = plots_data + _create_plots_data_from_control(
            name, durations, np.array(values), polar, unit_symbol, two_pi_factor
        )

    # Get the appropriate scaling for the time axis based on the total durations of all controls.
    time_scaling, time_prefix = get_units(
        [sum(plot_data.xdata) for plot_data in plots_data]
    )

    axes_list = create_axes(
        figure, 1, len(plots_data), width=FIG_WIDTH, height=FIG_WIDTH / 4, share_x=True
    ).flatten()

    if smooth:
        for axes, plot_data in zip(axes_list, plots_data):
            # Convert the list of durations into a list of times at the midpoints of each segment,
            # with two leading zero and two trailing total times.
            # Length of 'times' is m+4 (m is the number of segments).
            end_points = np.cumsum(plot_data.xdata)
            times = (
                np.concatenate(
                    [
                        [0.0, 0.0],
                        end_points - np.array(plot_data.xdata) * 0.5,
                        [end_points[-1], end_points[-1]],
                    ]
                )
                / time_scaling
            )
            # Pad each values array (1) with leading and trailing edge values,
            # to ensure smooth plot, and (2) with leading and trailing zeros,
            # to indicate that the pulse is zero outside the plot domain.
            # Length of 'values_arr' is m+4.
            values_arr = np.pad(
                np.pad(plot_data.values, ((1, 1)), "edge"), ((1, 1)), "constant"
            )

            axes.plot(times, values_arr, linewidth=2)
            axes.fill(times, values_arr, alpha=0.3)

            axes.axhline(y=0, linewidth=1, zorder=-1)
            axes.set_ylabel(plot_data.ylabel)

    else:
        for axes, plot_data in zip(axes_list, plots_data):
            # Convert the list of durations into times, including a leading zero. Length of 'times'
            # is m+1 (m is the number of segments).
            times = np.insert(np.cumsum(plot_data.xdata), 0, 0.0) / time_scaling
            # Pad each values array with leading and trailing zeros, to indicate that the pulse is
            # zero outside the plot domain. Length of 'values_arr' is m+2.
            values_arr = np.pad(plot_data.values, ((1, 1)), "constant")

            #              *---v2--*
            #              |       |
            #       *--v1--*       |        *-v4-*
            #       |              |        |    |
            #       |              *---v3---*    |
            # --v0--*                            *---v5--
            #       t0     t1      t2       t3   t4
            # To plot a piecewise-constant pulse, we need to sample from the times array at indices
            # [0, 0, 1, 1, ..., m-1, m-1, m, m  ], and from the values arrays at indices
            # [0, 1, 1, 2, ..., m-1, m,   m, m+1].
            time_indices = np.repeat(np.arange(len(times)), 2)
            value_indices = np.repeat(np.arange(len(values_arr)), 2)[1:-1]

            axes.plot(times[time_indices], values_arr[value_indices], linewidth=2)
            axes.fill(times[time_indices], values_arr[value_indices], alpha=0.3)

            axes.axhline(y=0, linewidth=1, zorder=-1)
            for time in times:
                axes.axvline(x=time, linestyle="--", linewidth=1, zorder=-1)

            axes.set_ylabel(plot_data.ylabel)

    axes_list[-1].set_xlabel(f"Time ({time_prefix}s)")


@qctrl_style()
@figure_as_kwarg_only
def plot_smooth_controls(
    controls: dict,
    polar: bool = True,
    unit_symbol: str = "Hz",
    two_pi_factor: bool = True,
    *,
    figure: plt.Figure,
):
    """
    Create a plot of the specified smooth controls.

    Parameters
    ----------
    controls : dict
        The dictionary of controls to plot. The keys should be the names of the controls, and the
        values represent the pulse by either (1) a dictionary with the 'times' and 'values' for
        that control, or (2) a list of samples, each a dictionary with 'time' and 'value' keys.
        The times must be in seconds and the values (possibly complex) in the units specified by
        `unit_symbol`.
        For example, the following two ``controls`` inputs would be valid (and equivalent)::

            controls={
              "Clock": {"times": [0.0, 0.5, 1.5, 2.0], "values": [0.1, 0.3, 0.2, 0.4]},
              "Microwave": {"times": [0.0, 1.0, 2.0], "values": [0.5 + 0.5j, 0.3, 0.2 - 0.3j]},
            }

            controls={
              "Clock": [
                {"time": 0.0, "value": 0.1},
                {"time": 0.5, "value": 0.3},
                {"time": 1.5, "value": 0.2},
                {"time": 2.0, "value": 0.4},
              ],
              "Microwave": [
                {"time": 0.0, "value": 0.5 + 0.5j},
                {"time": 1.0, "value": 0.3},
                {"time": 2.0, "value": 0.2 - 0.3j},
              ],
            }

    polar : bool, optional
        The mode of the plot when the values appear to be complex numbers.
        Plot magnitude and angle in two figures if set to True, otherwise plot I and Q
        in two figures. Defaults to True.
    unit_symbol : str, optional
        The symbol of the unit to which the controls values correspond. Defaults to "Hz".
    two_pi_factor : bool, optional
        Whether the values of the controls should be divided by 2π in the plots.
        Defaults to True.
    figure : matplotlib.figure.Figure, optional
        A matplotlib Figure in which to place the plots.
        If passed, its dimensions and axes will be overridden.
    """

    plots_data: list[Any] = []
    for name, control in controls.items():
        # Process each control depending whether it's passed as a list of samples...
        if isinstance(control, list):
            times = []
            values = []
            for sample in control:
                times.append(sample["time"])
                values.append(sample["value"])
        # ...or a times/values dictionary.
        else:
            times = control["times"]
            values = control["values"]
        # Create plot data for the control.
        plots_data = plots_data + _create_plots_data_from_control(
            name, times, np.array(values), polar, unit_symbol, two_pi_factor
        )

    # Get the appropriate scaling for the time axis based on the total durations of all controls.
    time_scaling, time_prefix = get_units(
        [max(plot_data.xdata) for plot_data in plots_data]
    )

    axes_list = create_axes(
        figure, 1, len(plots_data), width=FIG_WIDTH, height=FIG_WIDTH / 4, share_x=True
    ).flatten()

    for axes, plot_data in zip(axes_list, plots_data):
        # Pad with leading and trailing zeros, to indicate that the pulse is zero outside the plot
        # domain.
        times = np.pad(plot_data.xdata, ((1, 1)), "edge") / time_scaling
        values_arr = np.pad(plot_data.values, ((1, 1)), "constant")

        axes.plot(times, values_arr, linewidth=2)
        axes.fill(times, values_arr, alpha=0.3)

        axes.axhline(y=0, linewidth=1, zorder=-1)

        axes.set_ylabel(plot_data.ylabel)

    axes_list[-1].set_xlabel(f"Time ({time_prefix}s)")


# Internal named tuple containing data required to draw a single plot. Note that xdata can represent
# either durations (of segments) or times (of samples), depending on whether the plot is for a
# piecewise-constant or smooth pulse.
_PlotData = namedtuple("_PlotData", ["ylabel", "xdata", "values"])


def _create_plots_data_from_control(
    name, xdata, values, polar, unit_symbol, two_pi_factor
):
    """
    Create a list of _PlotData objects for the given control data.
    """

    # Scale values (and name) if they're to be divided by 2π
    if two_pi_factor:
        scaled_values = values / (2 * np.pi)
        scaled_name = f"{name}$/2\\pi$"
    else:
        scaled_values = values
        scaled_name = name

    if not np.iscomplexobj(values):
        # Real control.
        prefix_scaling, prefix = get_units(scaled_values)
        return [
            _PlotData(
                xdata=xdata,
                values=scaled_values / prefix_scaling,
                ylabel=f"{scaled_name}\n({prefix}{unit_symbol})",
            )
        ]

    if polar:
        # Complex control, split into polar coordinates.
        prefix_scaling, prefix = get_units(scaled_values)
        return [
            _PlotData(
                xdata=xdata,
                values=np.abs(scaled_values) / prefix_scaling,
                ylabel=f"{scaled_name}\nModulus\n({prefix}{unit_symbol})",
            ),
            _PlotData(
                xdata=xdata, values=np.angle(values), ylabel=f"{name}\nAngle\n(rad)"
            ),
        ]

    # Complex control, split into rectangle coordinates.
    prefix_scaling_x, prefix_x = get_units(np.real(scaled_values))
    prefix_scaling_y, prefix_y = get_units(np.imag(scaled_values))
    return [
        _PlotData(
            xdata=xdata,
            values=np.real(scaled_values) / prefix_scaling_x,
            ylabel=f"{scaled_name}\nI\n({prefix_x}{unit_symbol})",
        ),
        _PlotData(
            xdata=xdata,
            values=np.imag(scaled_values) / prefix_scaling_y,
            ylabel=f"{scaled_name}\nQ\n({prefix_y}{unit_symbol})",
        ),
    ]


@qctrl_style()
@figure_as_kwarg_only
def plot_sequences(sequences: dict, *, figure: plt.Figure):
    """
    Create plot of dynamical decoupling sequences.

    Parameters
    ----------
    sequences : dict
        The dictionary of sequences to plot. Works the same as the dictionary for
        plot_controls, but takes 'offset' instead of 'duration' and 'rotation'
        instead of 'value'. Rotations can be around any axis in the XY plane.
        Information about this axis is encoded in the complex argument of the
        rotation. For example, a pi X-rotation is represented by the complex
        number 3.14+0.j, whereas a pi Y-rotation is 0.+3.14j. The argument of the
        complex number is plotted separately as the azimuthal angle.
        For example, the following ``sequences`` inputs would be valid ::

            sequences={
              'X': [
                {'offset': 0.0, 'rotation': 0.1},
                {'offset': 0.5, 'rotation': 0.3},
                {'offset': 1.5, 'rotation': 0.2},
                {'offset': 2.0, 'rotation': 0.4},
              ],
              'XY': [
                {'offset': 0.0, 'rotation': 0.5 + 0.5j},
                {'offset': 1.0, 'rotation': 0.3},
                {'offset': 2.0, 'rotation': 0.2 - 0.3j},
              ],
            }

    figure : matplotlib.figure.Figure, optional
        A matplotlib Figure in which to place the plots.
        If passed, its dimensions and axes will be overridden.
    """

    check_argument(
        isinstance(sequences, dict) and len(sequences) > 0,
        "Sequences must be a non-empty dictionary.",
        {"sequences": sequences},
    )
    plots_data: list[Any] = []
    for name, pulses in sequences.items():
        offsets = [pulse["offset"] for pulse in pulses]
        rotations = np.array([pulse["rotation"] for pulse in pulses])
        plots_data = plots_data + _create_plots_data_from_sequence(
            name, offsets, rotations
        )

    axes_list = create_axes(
        figure, 1, len(plots_data), width=FIG_WIDTH, height=FIG_WIDTH / 4, share_x=True
    ).flatten()

    for axes, plot_data in zip(axes_list, plots_data):
        # The plot_data.offsets array contains only the points where the
        # dynamical decoupling pulses occur. For plotting purposes, it is
        # necessary to have three points describing each instantaneous pulse:
        # one at zero before the pulse, one with the actual value of the
        # pulse, and a third one at zero just after the pulse. The following
        # function triples the number of points in the time array so that
        # the pulses can be drawn like that.
        times = np.repeat(plot_data.offsets, 3)
        time_scaling, time_prefix = get_units(times)
        times /= time_scaling

        # Besides three points for each pulse, two extra points have to be
        # added: one before all the pulses and one after all of them.
        # np.pad() adds these points, with the first one located at t=0,
        # and the line after it makes sure that
        # the distance between the last point and the last pulse is the same
        # as the distance between the first point and the first pulse. This
        # gives an overall symmetric look to the plot.
        times = np.pad(times, ((1, 1)), "constant")
        times[-1] = times[-2] + times[1]

        values = np.zeros(3 * len(plot_data.rotations))
        values[1::3] = plot_data.rotations

        values = np.pad(values, ((1, 1)), "constant")

        axes.plot(times, values, linewidth=2)

        axes.axhline(y=0, linewidth=1, zorder=-1)
        for time in plot_data.offsets:
            axes.axvline(x=time, linestyle="--", linewidth=1, zorder=-1)

        axes.set_ylabel(plot_data.label)

    axes_list[-1].set_xlabel(f"Time ({time_prefix}s)")


_PlotSeqData = namedtuple("_PlotSeqData", ["label", "offsets", "rotations"])


def _create_plots_data_from_sequence(name, offsets, rotations):
    """
    Create a list of _PlotSeqData objects for the given dynamical decoupling data.
    """
    if not np.iscomplexobj(rotations):
        return [
            _PlotSeqData(
                offsets=offsets, rotations=rotations, label=f"{name}\nrotations\n(rad)"
            )
        ]
    return [
        _PlotSeqData(
            offsets=offsets,
            rotations=np.abs(rotations),
            label=f"{name}\nrotations\n(rad)",
        ),
        _PlotSeqData(
            offsets=offsets,
            rotations=np.angle(rotations),
            label=f"{name}\nazimuthal angles\n(rad)",
        ),
    ]



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Bloch sphere visualization module
"""

import json
import uuid
from typing import Optional

import numpy as np
from qctrlcommons.preconditions import check_argument

PAULI_X = np.array([[0, 1], [1, 0]], dtype=complex)
PAULI_Y = np.array([[0, -1j], [1j, 0]], dtype=complex)
PAULI_Z = np.array([[1, 0], [0, -1]], dtype=complex)

PAULIS = [PAULI_X, PAULI_Y, PAULI_Z]


def display_bloch_sphere(
    states: np.ndarray,
    color: Optional[str] = None,
    width: int = 300,
    visualizer_js: Optional[str] = None,
):
    r"""
    Display a trajectory on the Bloch sphere using state vectors as input.

    This function requires IPython, and you must run it from a Jupyter
    notebook. By default, it requires an Internet connection to fetch the
    JavaScript library for the Bloch sphere visualizer, but you can also
    use it offline by making a backup copy of the JavaScript file in your
    local filesystem.

    This function plots trajectories on the surface of the Bloch sphere if
    you provide normalized state vectors. To display trajectories in the
    interior of the sphere, provide non-normalized state vectors, or use
    `display_bloch_sphere_from_density_matrices
    <display_bloch_sphere_from_density_matrices.html>`_ or
    `display_bloch_sphere_from_bloch_vectors
    <display_bloch_sphere_from_bloch_vectors.html>`_.

    Parameters
    ----------
    states : np.ndarray
        A trajectory of single-qubit states represented by state vectors.
        This array must have shape ``[T,2]``, where ``T`` is the number of
        state vectors in the trajectory.
    color : str, optional
        A string identifying the color of the trajectory. The string must
        be a color value accepted by CSS, such as a hexadecimal code like
        "#00FF00" or a color name like "green". The exact types of
        values accepted might depend on your browser, but you can find an
        official list of color values as part of the `CSS standard
        <https://www.w3.org/TR/css-color-3/#colorunits>`_. If you don't
        pass a string, the default behavior is to use the color value
        "#EB6467".
    width : int, optional
        The width of the Bloch sphere, in pixels. Its height has the same
        value as the width. Defaults to 300 pixels.
    visualizer_js : str, optional
        A string with the location of the JavaScript library for the Bloch
        sphere visualizer. It can be a URL or a path to a local file. If
        you don't pass a string, the function uses the default online
        version of the Q-CTRL Visualizer JavaScript package.

    See Also
    --------
    display_bloch_sphere_from_bloch_vectors,
    display_bloch_sphere_from_density_matrices

    Notes
    -----
    This function represents the trajectory of single-qubit states as
    points on the Bloch sphere. The points on the Bloch sphere have
    spherical coordinates :math:`(r, \theta, \phi)`, which (up to a global
    phase) correspond to each state vector
    :math:`\left| \psi \right\rangle` in the trajectory according to:

    .. math::
        \left| \psi \right\rangle = \sqrt{r} \cos(\theta/2)
        \left| 0 \right\rangle + \sqrt{r} e^{i \phi} \sin(\theta/2)
        \left| 1 \right\rangle.

    The Bloch sphere has unit radius. If the states that you provide are
    not normalized, then :math:`r \ne 1` and you can draw trajectories
    inside the Bloch sphere.

    .. rubric:: Styling

    This function displays HTML and JavaScript in your Jupyter notebook. It
    creates four HTML elements that you can style by defining the following
    CSS classes:

    qctrlvisualizer
        CSS class for the outer ``<div>`` that contains the Bloch sphere
        visualization, progress bar, and button. The div has inline styles
        ``display:flex``, ``flex-direction:column``, and
        ``align-items:center``, so if you want to use different values for
        these properties you must either strip them out or use the
        ``!important`` rule when defining your ``qctrlvisualizer`` class.

    qctrlvisualizer-wrapper
        CSS class for the ``<div>`` that contains the Bloch sphere
        visualization. The div has inline styles ``margin:0.5rem 0``,
        ``width``, and ``height`` (based on the `width` you pass).

    qctrlvisualizer-progress-bar
        CSS class for the range ``<input>`` representing the progress bar.
        The input has inline style ``margin:0.5rem 0``.

    qctrlvisualizer-button
        CSS class for the ``<button>`` representing the play/pause/replay
        button. The button has inline style ``margin:0.5rem 0``.

    qctrlvisualizer-button-play
        CSS class for the ``<button>`` when the button text is "Play".

    qctrlvisualizer-button-pause
        CSS class for the ``<button>`` when the button text is "Pause".

    qctrlvisualizer-button-replay
        CSS class for the ``<button>`` when the button text is "Replay".
    """
    density_matrices = (
        np.asarray(states)[..., None] @ np.asarray(states)[..., None, :].conj()
    )

    display_bloch_sphere_from_density_matrices(
        density_matrices=density_matrices,
        color=color,
        width=width,
        visualizer_js=visualizer_js,
    )


def display_bloch_sphere_from_density_matrices(
    density_matrices: np.ndarray,
    color: Optional[str] = None,
    width: int = 300,
    visualizer_js: Optional[str] = None,
):
    r"""
    Display a trajectory in the Bloch sphere from input density matrices.

    This function requires IPython, and you must run it from a Jupyter
    notebook. By default, it requires an Internet connection to fetch the
    JavaScript library for the Bloch sphere visualizer, but you can also
    use it offline by making a backup copy of the JavaScript file in your
    local filesystem.

    Parameters
    ----------
    density_matrices : np.ndarray
        A trajectory of single-qubit states represented by density
        matrices. This array must have shape ``[T,2,2]``, where ``T`` is
        the number of density matrices in the trajectory.
    color : str, optional
        A string identifying the color of the trajectory. The string must
        be a color value accepted by CSS, such as a hexadecimal code like
        "#00FF00" or a color name like "green". The exact types of
        values accepted might depend on your browser, but you can find an
        official list of color values as part of the `CSS standard
        <https://www.w3.org/TR/css-color-3/#colorunits>`_. If you don't
        pass a string, the default behavior is to use the color value
        "#EB6467".
    width : int, optional
        The width of the Bloch sphere, in pixels. Its height has the same
        value as the width. Defaults to 300 pixels.
    visualizer_js : str, optional
        A string with the location of the JavaScript library for the Bloch
        sphere visualizer. It can be a URL or a path to a local file. If
        you don't pass a string, the function uses the default online
        version of the Q-CTRL Visualizer JavaScript package.

    See Also
    --------
    display_bloch_sphere,
    display_bloch_sphere_from_bloch_vectors

    Notes
    -----
    This function represents the trajectory of single-qubit states as
    points inside or on the Bloch sphere. These points correspond to each
    input density matrix :math:`\rho`, according to the following equation:

    .. math::
        \rho = \frac{1}{2} I + \frac{1}{2} \left( b_x \sigma_x + b_y
        \sigma_y + b_z \sigma_z \right),

    where :math:`I` is the :math:`2 \times 2` identity matrix, and
    :math:`b_x`, :math:`b_y`, and :math:`b_z` are Cartesian coordinates on
    the Bloch sphere.

    As the Pauli matrices :math:`\sigma_k` have zero trace and their square
    is the identity matrix, the value of each Cartesian coordinate
    :math:`b_k` is:

    .. math::
        b_k = \mathrm{Tr} \left\{ \sigma_k \rho \right\},

    which is just the expectation value of :math:`\sigma_k`.

    .. rubric:: Styling

    This function displays HTML and JavaScript in your Jupyter notebook. It
    creates four HTML elements that you can style by defining the following
    CSS classes:

    qctrlvisualizer
        CSS class for the outer ``<div>`` that contains the Bloch sphere
        visualization, progress bar, and button. The div has inline styles
        ``display:flex``, ``flex-direction:column``, and
        ``align-items:center``, so if you want to use different values for
        these properties you must either strip them out or use the
        ``!important`` rule when defining your ``qctrlvisualizer`` class.

    qctrlvisualizer-wrapper
        CSS class for the ``<div>`` that contains the Bloch sphere
        visualization. The div has inline styles ``margin:0.5rem 0``,
        ``width``, and ``height`` (based on the `width` you pass).

    qctrlvisualizer-progress-bar
        CSS class for the range ``<input>`` representing the progress bar.
        The input has inline style ``margin:0.5rem 0``.

    qctrlvisualizer-button
        CSS class for the ``<button>`` representing the play/pause/replay
        button. The button has inline style ``margin:0.5rem 0``.

    qctrlvisualizer-button-play
        CSS class for the ``<button>`` when the button text is "Play".

    qctrlvisualizer-button-pause
        CSS class for the ``<button>`` when the button text is "Pause".

    qctrlvisualizer-button-replay
        CSS class for the ``<button>`` when the button text is "Replay".
    """
    bloch_vectors = _expectation_values(
        np.asarray(density_matrices), np.asarray(PAULIS)
    )

    display_bloch_sphere_from_bloch_vectors(
        bloch_vectors=bloch_vectors,
        color=color,
        width=width,
        visualizer_js=visualizer_js,
    )


def _expectation_values(
    density_matrices: np.ndarray, observables: np.ndarray
) -> np.ndarray:
    """
    Return expectation values of `observables` for `density_matrices`.

    Parameters
    ----------
    density_matrices : np.ndarray
        A batch of :math:`M` density matrices of shape :math:`D \times D`,
        with respect to which you want to calculate the expectation values.
    observables : np.ndarray
        A batch of :math:`N` matrices of shape :math:`D \times D`, whose
        expectation values you want to calculate.

    Returns
    -------
    np.ndarray
        An array of shape ``[M,N]`` with the real expectation values.
    """
    return np.real(
        np.trace(
            np.matmul(observables[None, ...], density_matrices[:, None, ...]),
            axis1=-2,
            axis2=-1,
        )
    )


def display_bloch_sphere_from_bloch_vectors(
    bloch_vectors: np.ndarray,
    color: Optional[str] = None,
    width: int = 300,
    visualizer_js: Optional[str] = None,
):
    r"""
    Display a Bloch sphere with an animation of the Bloch vectors.

    This function requires IPython, and you must run it from a Jupyter
    notebook. By default, it requires an Internet connection to fetch the
    JavaScript library for the Bloch sphere visualizer, but you can also
    use it offline by making a backup copy of the JavaScript file in your
    local filesystem.

    Parameters
    ----------
    bloch_vectors : np.ndarray
        A trajectory of Bloch vectors (that is, points either on or inside
        the Bloch sphere) given in Cartesian coordinates. This array must
        have shape ``[T,3]``, where ``T`` is the number of vectors in the
        trajectory.
    color : str, optional
        A string identifying the color of the trajectory. The string must
        be a color value accepted by CSS, such as a hexadecimal code like
        "#00FF00" or a color name like "green". The exact types of
        values accepted might depend on your browser, but you can find an
        official list of color values as part of the `CSS standard
        <https://www.w3.org/TR/css-color-3/#colorunits>`_. If you don't
        pass a string, the default behavior is to use the color value
        "#EB6467".
    width : int, optional
        The width of the Bloch sphere, in pixels. Its height has the same
        value as the width. Defaults to 300 pixels.
    visualizer_js : str, optional
        A string with the location of the JavaScript library for the Bloch
        sphere visualizer. It can be a URL or a path to a local file. If
        you don't pass a string, the function uses the default online
        version of the Q-CTRL Visualizer JavaScript package.

    Raises
    ------
    QctrlArgumentsValueError
        If `bloch_vectors` is of complex type.

    See Also
    --------
    display_bloch_sphere,
    display_bloch_sphere_from_density_matrices

    Notes
    -----
    The Bloch vector :math:`\mathbf{b}` is a 3D real vector that represents
    a point either on or inside the Bloch sphere. It is equivalent to a
    single-qubit state represented by a density matrix :math:`\rho`,
    according to the following equation:

    .. math::
        \rho = \frac{1}{2} I + \frac{1}{2} \left( b_x \sigma_x + b_y
        \sigma_y + b_z \sigma_z \right),

    where :math:`I` is the :math:`2 \times 2` identity matrix, and
    :math:`b_x`, :math:`b_y`, and :math:`b_z` are the components of the
    Bloch vector in Cartesian coordinates.

    For points on the surface of the Bloch sphere (that is, when the norm
    of the vector has value 1), the Bloch vector represents a pure state.
    In this case, if you write the Bloch vector in terms of the spherical
    coordinates :math:`(\theta, \phi)`, such that
    :math:`b_x = \sin\theta \cos\phi`, :math:`b_y = \sin\theta \sin\phi`,
    and :math:`b_z = \cos \theta`, then the state vector it represents is:

    .. math::
        \left| \psi \right\rangle = \cos(\theta/2) \left| 0 \right\rangle
        + e^{i \phi} \sin(\theta/2) \left| 1 \right\rangle.

    .. rubric:: Styling

    This function displays HTML and JavaScript in your Jupyter notebook. It
    creates four HTML elements that you can style by defining the following
    CSS classes:

    qctrlvisualizer
        CSS class for the outer ``<div>`` that contains the Bloch sphere
        visualization, progress bar, and button. The div has inline styles
        ``display:flex``, ``flex-direction:column``, and
        ``align-items:center``, so if you want to use different values for
        these properties you must either strip them out or use the
        ``!important`` rule when defining your ``qctrlvisualizer`` class.

    qctrlvisualizer-wrapper
        CSS class for the ``<div>`` that contains the Bloch sphere
        visualization. The div has inline styles ``margin:0.5rem 0``,
        ``width``, and ``height`` (based on the `width` you pass).

    qctrlvisualizer-progress-bar
        CSS class for the range ``<input>`` representing the progress bar.
        The input has inline style ``margin:0.5rem 0``.

    qctrlvisualizer-button
        CSS class for the ``<button>`` representing the play/pause/replay
        button. The button has inline style ``margin:0.5rem 0``.

    qctrlvisualizer-button-play
        CSS class for the ``<button>`` when the button text is "Play".

    qctrlvisualizer-button-pause
        CSS class for the ``<button>`` when the button text is "Pause".

    qctrlvisualizer-button-replay
        CSS class for the ``<button>`` when the button text is "Replay".
    """
    check_argument(
        not np.iscomplexobj(bloch_vectors),
        "The Bloch vectors must not have complex type.",
        {"bloch_vectors": bloch_vectors},
    )

    identifier = uuid.uuid4().hex

    visualization_data = {
        "data": {
            "qubits": [
                {
                    "name": "qubit1",
                    "x": [vector[0] for vector in bloch_vectors],
                    "y": [vector[1] for vector in bloch_vectors],
                    "z": [vector[2] for vector in bloch_vectors],
                }
            ]
        }
    }

    color = color or "#EB6467"

    _display_visualization(
        visualization_data=visualization_data,
        identifier=identifier,
        color=color,
        width=width,
        height=width,
        visualizer_js=visualizer_js,
    )


def _display_visualization(
    visualization_data: dict,
    identifier: str,
    width: int,
    height: int,
    color: str,
    visualizer_js: Optional[str] = None,
):
    """
    Call the Q-CTRL Visualizer JavaScript to display the Bloch sphere.

    You can find the Q-CTRL Visualizer JavaScript package at
    https://www.npmjs.com/package/@qctrl/visualizer .

    Parameters
    ----------
    visualization_data : dict
        A dictionary containing serialized data that is understood by the
        `Visualizer` class of the ``Visualizer.js`` script.
    identifier : str
        A string appended to the end of the ``id`` of the HTML objects that
        this function creates.
    width : int
        The width of the visualization, in pixels.
    height : int
        The height of the visualization, in pixels.
    color : str
        A string identifying the color of the trajectory on the Bloch
        sphere. The string must be a color value accepted by CSS, such as a
        hexadecimal code like "#00FF00" or a color name like
        "green".
    visualizer_js : str, optional
        A string with the location of the JavaScript library for the Bloch
        sphere visualizer. It can be a URL or a path to a local file. If
        you don't pass a string, the function uses the default online
        version of the Q-CTRL Visualizer JavaScript package.
    """
    # pylint: disable=import-error,import-outside-toplevel
    from IPython.display import (
        HTML,
        Javascript,
        display,
    )

    theme_settings = {"highlightColor": color, "pathColor": color}

    labels = {
        "xAxis": True,
        "yAxis": True,
        "zAxis": True,
        "theta": True,
        "phi": True,
        "northPole": True,
        "southPole": True,
        # Prevent "Non Error Trajectory" dot from appearing.
        "nonErrorState": False,
    }

    default_url = (
        "https://cdn.jsdelivr.net/npm/@qctrl/visualizer@3.1.14/umd/visualizer.min.js"
    )

    # If no custom Visualizer.js is provided, use the default one.
    visualizer_js = visualizer_js or default_url

    html_code = f"""
    <div
        class="qctrlvisualizer"
        style="display:flex;flex-direction:column;align-items:center;"
    >
        <div
            class="qctrlvisualizer-wrapper"
            style="width:{width}px;height:{height}px;margin:0.5rem 0"
            id="qctrlvisualizer-wrapper-{identifier}"
        ></div>
        <input
            class="qctrlvisualizer-progress-bar"
            style="margin:0.5rem 0"
            id="qctrlvisualizer-progress-bar-{identifier}"
            type="range"
            min="0"
            max="1"
            step="0.01"
            value="0"
        >
        <button
            class="qctrlvisualizer-button qctrlvisualizer-button-play"
            style="margin:0.5rem 0"
            id="qctrlvisualizer-button-{identifier}"
        >Play</button>
    </div>
    """

    js_code = f"""
    let isPlaying = false;
    let progress = 0;

    const visualizationData = {json.dumps(visualization_data)};
    const themeSettings = {json.dumps(theme_settings)};
    const labels = {json.dumps(labels)};

    const wrapper = document.getElementById("qctrlvisualizer-wrapper-{identifier}");
    const progressBar = document.getElementById("qctrlvisualizer-progress-bar-{identifier}");
    const button = document.getElementById("qctrlvisualizer-button-{identifier}");

    function updateButton () {{
      button.classList.remove(...[
        "qctrlvisualizer-button-play",
        "qctrlvisualizer-button-pause",
        "qctrlvisualizer-button-replay"
      ]);
      if (isPlaying) {{
        button.classList.add("qctrlvisualizer-button-pause");
        button.innerHTML = "Pause";
        return;
      }}
      if (progress>=1) {{
        button.classList.add("qctrlvisualizer-button-replay");
        button.innerHTML = "Replay";
        return;
      }}
      button.classList.add("qctrlvisualizer-button-play");
      button.innerHTML = "Play";
    }}

    button.onclick = () => {{
      isPlaying = !isPlaying;
      if (progress >= 1) progress = 0;
      updateButton();
      visualizer.update({{ isPlaying, progress }});
    }};

    progressBar.oninput = ({{ target }}) => {{
      progress = +target.value;
      visualizer.update({{ progress }});
      updateButton();
    }};

    const onUpdate = ({{ target, data }}) => {{
      progress = data.progress;
      progressBar.value = progress;
      if (progress >= 1) {{
        isPlaying = false;
        target.update({{ isPlaying }});
        updateButton();
      }}
    }};

    const visualizer = new Visualizer({{
      visualizationData,
      wrapper,
      onUpdate,
      labels,
    }}).init();

    visualizer.update({{ themeSettings }});
    """

    custom_js_loader = f"""
    function visualizerLoader() {{
      if ((typeof Visualizer) == "function") {{
        console.log("Using preloaded Q-CTRL Visualizer JavaScript package.");
        displayBlochSphere();
        return;
      }}

      try {{
        console.log("Attempting to load {visualizer_js} with require.js.");
        requirejs(["{visualizer_js}"], displayBlochSphere, displayErrorMessage);
      }} catch(error) {{
        var existing_script = document.getElementById("qctrlvisualizer-script");

        if (existing_script !== null) {{
          console.log(
            "Script tag for the Q-CTRL Visualizer JavaScript package already exists."
            + " Delaying execution of function until script is loaded."
          );
          existing_script.addEventListener("load", displayBlochSphere);
          existing_script.addEventListener("error", displayErrorMessage);
          return;
        }}

        console.log("Attempting to load {visualizer_js} with script tag.");
        var script = document.createElement("script");
        script.onload = displayBlochSphere;
        script.onerror = displayErrorMessage;
        script.id = "qctrlvisualizer-script";
        script.src = "{visualizer_js}";
        document.head.appendChild(script);
      }}
    }}

    function displayBlochSphere() {{ {js_code} }}

    function displayErrorMessage() {{
      console.log("Failed to load {visualizer_js}.");
      const wrapper = document.getElementById("qctrlvisualizer-wrapper-{identifier}");
      wrapper.innerHTML = "Could not load JavaScript at {visualizer_js}.";
    }}

    visualizerLoader();
    """

    display(HTML(html_code))
    display(Javascript(custom_js_loader))



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Functions for plotting cost vs iterations.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import (
    LinearSegmentedColormap,
    TwoSlopeNorm,
)
from qctrlcommons.preconditions import check_argument

from .style import (
    FIG_HEIGHT,
    FIG_WIDTH,
    QCTRL_STYLE_COLORS,
    qctrl_style,
)
from .utils import figure_as_kwarg_only


@qctrl_style()
@figure_as_kwarg_only
def plot_wigner_function(
    wigner_function: np.ndarray,
    position: np.ndarray,
    momentum: np.ndarray,
    contour_count: int = 100,
    *,
    figure: plt.Figure,
):
    """
    Create a contour plot of the specified Wigner function.

    Parameters
    ----------
    wigner_function : np.ndarray
        The Wigner function values. Only the real part of this array will be plotted.
        Must be a 2D array of shape ``(L, K)``.
    position : np.ndarray
        The dimensionless position vector.
        Must be a strictly increasing 1D array of length `L`.
    momentum : np.ndarray
        The dimensionless momentum vector.
        Must be a strictly increasing 1D array of length `K`.
    contour_count : int, optional
        The number of contour lines, the larger the value the finer the contour plot will be.
        Defaults to 100.
    figure : matplotlib.figure.Figure, optional
        A matplotlib Figure in which to place the plots.
        If passed, its dimensions and axes will be overridden.
    """
    check_argument(
        isinstance(wigner_function, np.ndarray) and wigner_function.ndim == 2,
        "The Wigner function must be in a 2D array.",
        {"wigner_function": wigner_function},
    )

    _check_axis_array(position, "position", wigner_function, 0)
    _check_axis_array(momentum, "momentum", wigner_function, 1)

    check_argument(
        isinstance(contour_count, (int, np.integer)) and contour_count > 0,
        "The number of contours must be a positive integer.",
        {"contour_count": contour_count},
    )

    figure.set_figwidth(FIG_WIDTH)
    figure.set_figheight(FIG_HEIGHT)

    axes = figure.subplots(nrows=1, ncols=1)
    axes.set_aspect("equal")

    cmap = _get_cmap()

    wigner_max_abs = np.max(np.abs(wigner_function))
    norm = TwoSlopeNorm(vcenter=0.0, vmin=-wigner_max_abs, vmax=wigner_max_abs)

    # The Wigner function is a real function, but we only plot the real part
    # in case the user passes a complex array.
    # We also transpose the array so that position is on the x-axis.
    wigner_plot = axes.contourf(
        position,
        momentum,
        wigner_function.T.real,
        levels=contour_count,
        cmap=cmap,
        norm=norm,
    )

    figure.colorbar(wigner_plot, ax=axes)

    axes.set_xlabel("Position")
    axes.set_ylabel("Momentum")


def _check_axis_array(array, name, wigner_function, axis):
    check_argument(
        isinstance(array, np.ndarray) and array.ndim == 1,
        f"The {name} must be in a 1D array.",
        {name: array},
    )

    check_argument(
        all(np.diff(array) > 0),
        f"The {name} must be a strictly increasing array.",
        {name: array},
    )

    rows_or_columns = ["rows", "columns"]

    check_argument(
        wigner_function.shape[axis] == len(array),
        f"The length of the {name} array must match the number of {rows_or_columns[axis]} in the"
        " Wigner function array.",
        {"wigner_function": wigner_function, name: array},
        extras={
            "wigner_function.shape": wigner_function.shape,
            f"len({name})": len(array),
        },
    )


def _get_cmap():
    def hex_to_rgb(hex_input):
        return [int(hex_input[i : i + 2], 16) / 255 for i in [1, 3, 5]]

    red = QCTRL_STYLE_COLORS[1]
    white = "#ffffff"
    blue = QCTRL_STYLE_COLORS[5]
    return LinearSegmentedColormap.from_list(
        name="qctrl_cmap", colors=[hex_to_rgb(c) for c in [red, white, blue]]
    )
